// (C) 2017-2020 by folkert van heusden, released under AGPL v3.0
#include <assert.h>
#include <string>
#include <string.h>
#include <unistd.h>

#include <curl/curl.h>

#include "utils.h"
#include "source.h"
#include "source_gstreamer.h"
#include "picio.h"
#include "error.h"
#include "log.h"

source_gstreamer::source_gstreamer(const std::string & id, const std::string & descr, const std::string & exec_failure, const std::string & host, const int port, const double max_fps, resize *const r, const int resize_w, const int resize_h, const int loglevel, const double timeout, std::vector<filter *> *const filters) : source(id, descr, exec_failure, max_fps, r, resize_w, resize_h, loglevel, timeout, filters), host(host), port(port)
{
}

source_gstreamer::~source_gstreamer()
{
	stop();
}

void source_gstreamer::operator()()
{
	log(id, LL_INFO, "source gstreamer thread started (%s:%d)", host.c_str(), port);

	set_thread_name("src_gstreamer");

	char *buffer = nullptr;
	size_t buffer_len = 0, buffer_offset = 0, segment_len = 0;

	int fd = -1;
	bool header = true, boundary_unknown = true;

	char boundary[256] = { 0 };
	size_t boundary_len = 0;

	for(;!local_stop_flag;)
	{
		if (fd == -1) {
			myusleep(101000);

			fd = connect_to(host, port, &local_stop_flag);

			header = true;

			free(buffer);
			buffer = nullptr;

			buffer_len = buffer_offset = 0;
		}

		if (fd == -1)
			continue;

		if (buffer_len - buffer_offset < 4096) {
			buffer = (char *)realloc(buffer, buffer_len + 16384 + 1);
			if (!buffer) {
				free(buffer);
				log(id, LL_ERR, "cannot allocate 4kB");
				break;
			}

			buffer_len += 16384;
		}

		fd_set rfds;
		FD_ZERO(&rfds);
		FD_SET(fd, &rfds);

		struct timeval tv;
		tv.tv_sec  = 0;
		tv.tv_usec = 100 * 1000;

		/* wait for connection */
		int s_rc = select(fd + 1, &rfds, NULL, NULL, &tv);
		if (s_rc == 0)	// time out
			continue;

		if (s_rc == -1)	// error
		{
			log(LL_ERR, "Select failed during read from %s:%d: %s", host.c_str(), port, strerror(errno));

			close(fd);
			fd = -1;
			continue;
		}

		if (!FD_ISSET(fd, &rfds)) {
			log(id, LL_DEBUG, "read-socket has no data?!");
			continue;
		}

		int rc = read(fd, &buffer[buffer_offset], 4096);
		if (rc <= 0) {
			log(id, LL_ERR, "cannot read from socket: %s", strerror(errno));

			close(fd);
			fd = -1;

			continue;
		}

		buffer_offset += rc;
		buffer[buffer_offset] = 0x00;

		if (header) {
			char *endp = strstr(buffer, "\r\n\r\n");

//			printf("%d,%d | %zu |%d,%d,%d|%zu>%s|%*s\n", header, boundary_unknown, buffer_offset, !!endp, buffer_offset > boundary_len, memcmp(buffer, boundary, boundary_len), boundary_len, boundary, int(boundary_len), buffer);

			if (endp) {
				if (boundary_unknown || (buffer_offset > boundary_len && memcmp(buffer + 2, boundary, boundary_len) == 0)) {
					char *content_len = strstr(buffer, "Content-Length:");
					if (!content_len) {
						log(id, LL_ERR, "invalid header (1)");
						close(fd);
						fd = -1;
						continue;
					}
					
					if (boundary_unknown) {
						boundary_unknown = false;
						char *be = strstr(buffer, "\r\n");
						if (!be) {
							log(id, LL_ERR, "invalid header (2)");
							close(fd);
							fd = -1;
							continue;
						}

						boundary_len = std::min((long int)(be - buffer), (long int)(sizeof boundary - 1));
						strncpy(boundary, buffer, boundary_len);
						boundary[boundary_len] = 0x00;
					}

					*endp = 0x00;
//					printf("HEADER: |%s| %zu\n", buffer, strlen(buffer) + 4);

					segment_len = atoi(&content_len[16]);
					//printf("len %zu\n", segment_len);
					//printf("boundary |%s|\n", boundary);

					size_t left = buffer_offset - ((endp - buffer) + 4);
//printf("head should be %zu bytes (offset %zu)\n", left, buffer_offset);
//printf("after header: %s\n", endp + 4);
					if (left > 0) {
						//printf("---\n");
						//printf("%s\n", buffer);
						//printf("===\n");
						//printf("%s\n", endp + 4);
						//printf("---\n");
						memmove(&buffer[0], endp + 4, left);
						buffer_offset = left;
						buffer[left] = 0x00;
						//printf("left (%zu): %s\n", left, buffer);
					}
					else if (left < 0) {
						log(id, LL_ERR, "source gstreamer internal error (1): %d is < 0", left);
						buffer_offset = 0;
					}
					else {
						buffer_offset = 0;
					}

					header = false;
	//printf("* end header *\n");
				}
			}
		}
		else if (buffer_offset >= segment_len) {
			// FIXME resize
			//printf("frame %zu (%zu)\n", segment_len, buffer_offset);
			//printf("frame %s\n", buffer);
			set_frame(E_JPEG, (const uint8_t *)buffer, segment_len);

			size_t left = buffer_offset - segment_len;
			if (left > 0) {
				memmove(&buffer[0], &buffer[segment_len], left);
			buffer[left] = 0x00;
			//printf("frame left |%s|\n", buffer);
				buffer_offset = left;
			}
			else if (left < 0) {
				log(id, LL_ERR, "source gstreamer internal error (2): %d is < 0", left);
				buffer_offset = 0;
			}
			else {
				buffer_offset = 0;
			}

			header = true;
		}
	}

	if (fd != -1)
		close(fd);

	free(buffer);

	log(id, LL_INFO, "source gstreamer thread terminating");
}
